﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Commander.API;
using Commander.Core;
using Commander.Core.Command;
using Commander.Dialogs;
using Commander.UI;
using Commander.UI.Presenters;
using Microsoft.WindowsAPICodePack.Taskbar;

namespace Commander
{
	public partial class MainForm : Form, ICommandTabHost, ICommandAccess
	{
		private const string AppId = "_Commander";

		private readonly ApplicationSettingsManager _applicationSettingsManager;
		private TabSettings _tabSettings;
		private MouseButtons _tabContainersLastMouseAction = MouseButtons.None;
		private readonly CommandHistoryManager _commandHistoryManager;
		private JumpListCustomCategory _commandJumpListCategory;
		private readonly AutoCompletePresenter _autoCompletePresenter;
		private readonly ICommandManager _internalCommandManager;
		private readonly ClipboardHistoryPresenter _clipboardHistoryPresenter;
		private readonly IClipboardHistory _clipboardHistory;
		private readonly CommandLoader _commandLoader;
		private CommandHost _commandHost;
		private CommandLoadStatusViewPresenter _commandLoadStatusViewPresenter;

		// Used for toggling between fullscreen and "normal" view
		private FormBorderStyle _normalBorderStyle = FormBorderStyle.Sizable;
		private Point _normalLocation;
		private Size _normalSize;
		private FormWindowState _prevWindowState;

		public MainForm()
        {
            InitializeComponent();

			_internalCommandManager = new CommandManager(this);
			_commandLoader = new CommandLoader(new DirectoryInfo(Path.Combine(Application.StartupPath, "Commands")));

			var historyPath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "Commander");
			historyPath = Path.Combine(historyPath, "history.txt");
			_commandHistoryManager = new CommandHistoryManager(new FileInfo(historyPath));
        	_applicationSettingsManager = new ApplicationSettingsManager();
			_autoCompletePresenter = new AutoCompletePresenter(this, _commandHistoryManager);
			_clipboardHistory = new ClipboardHistoryManager();
			_clipboardHistoryPresenter = new ClipboardHistoryPresenter(_clipboardHistory, this, this);

			if (TaskbarManager.IsPlatformSupported)
			{
				TaskbarManager.Instance.ApplicationId = AppId;
			}
        }

		private void NewCommandTab(ShellTypes type = ShellTypes.Cmd)
		{
			if (InvokeRequired)
			{
				Invoke(new Action<ShellTypes>(NewCommandTab), type);
				return;
			}

			string prompt = String.Format("{0}@{1}:~$ ", Environment.UserName, Environment.MachineName);
			var tab = new ShellTabPresenter(this, _internalCommandManager, _commandHistoryManager,
											_clipboardHistory, _autoCompletePresenter, 
											Encoding.GetEncoding(850), prompt, _tabSettings, type);
			_commandTabContainer.SelectedTab = tab.View;

			tab.Init();
			CreateShellThumbnail(tab.View);
		}

		public void AddTab(TabPage tab)
		{
			_commandTabContainer.TabPages.Add(tab);
		}


		#region ICommandAccess Implementation
		public void AppendOutput(string output, bool newLine, bool addPrompt)
		{
			var activeTab = GetActiveTab();
			activeTab.AppendOutput(output, newLine, addPrompt);
		}

		public void ClearOutput()
		{
			var activeTab = GetActiveTab();
			activeTab.ClearOutput();
		}

		public void SetDirectory(string dir)
		{
			var activeTab = GetActiveTab();
			activeTab.SetDirectory(dir);
		}

		public string WorkingDirectory
		{
			get
			{
				var activeTab = GetActiveTab();
				return activeTab.WorkingDirectory;
			}
		}

		public ICommandHistory History
		{
			get { return _commandHistoryManager; }
		}

		public void CloseInstance()
		{
			CloseCurrentTab();
		}
		#endregion


		protected override void OnShown(EventArgs e)
		{
			base.OnShown(e);

			_tabSettings = _applicationSettingsManager.LoadTabSettings();

			CreateTaskbarTasks();

			_commandHost = new CommandHost(_internalCommandManager, _commandLoader);
			_commandLoadStatusViewPresenter = new CommandLoadStatusViewPresenter(statusBar, _commandHost, tslCommandsStatus);
			_commandLoadStatusViewPresenter.Init();

			NewCommandTab();

			_commandHistoryManager.Init();
			_autoCompletePresenter.Init();
			_autoCompletePresenter.ItemClicked += (sender, commandArg) => {
				_autoCompletePresenter.Hide();

			    var tabPresenter = GetActiveTab();
			    tabPresenter.InsertCommand(commandArg.Command);
				tabPresenter.Focus();
			};

			_clipboardHistoryPresenter.Init();
		}


		protected override void OnClientSizeChanged(EventArgs e)
		{
			_prevWindowState = WindowState;

			base.OnClientSizeChanged(e);
		}
		

		protected override void OnFormClosing(FormClosingEventArgs e)
		{
			base.OnFormClosing(e);

			CloseAllTabs();
		}

		protected override void OnMouseWheel(MouseEventArgs e)
		{
			base.OnMouseWheel(e);
			SwitchTab(e.Delta < 0 ? -1 : 1);
		}

		protected override void OnKeyUp(KeyEventArgs e)
		{
			base.OnKeyUp(e);

			if (e.Control)
			{
				if (e.KeyCode == Keys.Tab)
				{
					int direction = (e.Shift ? -1 : 1);
					SwitchTab(direction);
				}
			}
		}

		private void SwitchTab(int direction)
		{
			if (_commandTabContainer.TabCount < 1)
			{
				return;
			}

			int current = _commandTabContainer.SelectedIndex += direction;
			if (current >= _commandTabContainer.TabCount)
			{
				current = 0;
			}

			if (current < 0)
			{
				current = _commandTabContainer.TabCount - 1;
			}

			_commandTabContainer.SelectedIndex = current;
			var tab = _commandTabContainer.SelectedTab;
			TaskbarManager.Instance.TabbedThumbnail.SetActiveTab(tab);
		}

		private void CreateTaskbarTasks()
		{
			JumpList jumpList = JumpList.CreateJumpList();
			_commandJumpListCategory = new JumpListCustomCategory("Commands");
			jumpList.AddCustomCategories(_commandJumpListCategory);
			JumpListTask task = new JumpListLink(String.Format("{0}\\Commander.exe new-tab", Application.StartupPath), "New Tab");
			jumpList.AddUserTasks(task);
			jumpList.Refresh();
		}

		private void CreateShellThumbnail(TabPage tab)
		{
			if (!TaskbarManager.IsPlatformSupported)
			{
				return;
			}

			var thumbnailPreview = new TabbedThumbnail(Handle, tab) {
			    Title = tab.Text,
			    Tooltip = tab.Text,
			    DisplayFrameAroundBitmap = false
			};

			thumbnailPreview.SetWindowIcon(Icon);
			thumbnailPreview.SetImage(TabbedThumbnailScreenCapture.GrabWindowBitmap(tab.Handle, tab.Size));
			thumbnailPreview.TabbedThumbnailClosed += (sender, e) => CloseTab(tab);
			thumbnailPreview.TabbedThumbnailBitmapRequested	+= TabThumbnailBitmapRequested;
			thumbnailPreview.TabbedThumbnailActivated += TabbedThumbnailActivated;
			
			TaskbarManager.Instance.TabbedThumbnail.AddThumbnailPreview(thumbnailPreview);
			TaskbarManager.Instance.TabbedThumbnail.SetActiveTab(tab);
			TaskbarManager.Instance.TabbedThumbnail.InvalidateThumbnails();
		}

		private void TabbedThumbnailActivated(object sender, TabbedThumbnailEventArgs e)
		{
			TabPage tabPage = (TabPage)FromHandle(e.WindowHandle);
			Debug.WriteLine("Tab page activated: " + tabPage.Text);

			_commandTabContainer.SelectedTab = tabPage;
			if (WindowState == FormWindowState.Minimized)
			{
				WindowState = _prevWindowState;
			}
		}

		private void TabThumbnailBitmapRequested(object sender, TabbedThumbnailBitmapRequestedEventArgs e)
		{
			var thumbnail = (TabbedThumbnail)sender;
			Debug.WriteLine("TabThumbnail Requested for tab " + thumbnail.Title);

			thumbnail.InvalidatePreview();
			Bitmap bmp = null;
			try
			{
				var size = thumbnail.ClippingRectangle.HasValue ? thumbnail.ClippingRectangle.Value.Size : _commandTabContainer.Size;
				bmp = TabbedThumbnailScreenCapture.GrabWindowBitmap(e.WindowHandle, size);

				if (bmp == null)
				{
					return;
				}

				thumbnail.SetImage(bmp);
			}
			catch { }
			finally
			{
				if (bmp != null)
				{
					bmp.Dispose();
				}
			}
		}

		private static void UpdateShellThumbnail(TabPage tab)
		{
			var thumbnail = GetThumbnailForTab(tab);
			if (thumbnail == null)
			{
				return;
			}

			thumbnail.Title = tab.Text;
			thumbnail.Tooltip = tab.Text;

			var windowBitmap = TabbedThumbnailScreenCapture.GrabWindowBitmap(tab.Handle, tab.Size);
			thumbnail.SetImage(windowBitmap);
		}

		private static TabbedThumbnail GetThumbnailForTab(TabPage tab)
		{
			if (!TaskbarManager.IsPlatformSupported)
			{
				return null;
			}

			return TaskbarManager.Instance.TabbedThumbnail.GetThumbnailPreview(tab);
		}

		private void CloseCurrentTab()
		{
			if (_commandTabContainer.SelectedTab == null)
			{
				return;
			}

			CloseTab(_commandTabContainer.SelectedIndex);
		}

		public void CloseTab(TabPage tab)
		{
			int index = _commandTabContainer.TabPages.IndexOf(tab);
			CloseTab(index);
		}

		public void CloseTab(int index)
		{
			CloseTab(index, true);
		}

		public void CloseTab(int index, bool exitOnLastTabClosed)
		{
			if (index < 0 || index > _commandTabContainer.TabCount - 1)
			{
				return;
			}

			TabPage tab = _commandTabContainer.TabPages[index];
			_commandTabContainer.TabPages.Remove(tab);
			
			if (index > 0)
			{
				_commandTabContainer.SelectedTab = _commandTabContainer.TabPages[--index];
			}

			if (_commandTabContainer.TabPages.Count < 1 && exitOnLastTabClosed)
			{
				Close();
				return;
			}
		}

		public void CloseAllTabs()
		{
			_commandTabContainer.SuspendLayout();
			while (_commandTabContainer.TabCount > 0)
			{
				CloseTab(0, false);
			}
			_commandTabContainer.ResumeLayout(true);
		}

		private void RenameTab(TabPage tab)
		{
			using (var dlg = new RenameDialog())
			{
				dlg.NameText = tab.Text;
				if (dlg.ShowDialog(this) == DialogResult.OK)
				{
					tab.Text = dlg.NameText;
				}
			}

			UpdateShellThumbnail(tab);
		}

		private ShellTabPresenter GetActiveTab()
		{
			if (_commandTabContainer.SelectedTab == null)
			{
				return null;
			}

			return _commandTabContainer.SelectedTab.Tag as ShellTabPresenter;
		}

		public void Copy()
		{
			var activeTab = GetActiveTab();
			activeTab.Copy();
			_clipboardHistory.Add(Clipboard.GetText());
		}

		public void Paste()
		{
			var activeTab = GetActiveTab();
			activeTab.Paste();
		}

		public void Paste(string content)
		{
			var activeTab = GetActiveTab();
			activeTab.Paste(content);
		}

		private void OnNewTabButtonClicked(object sender, EventArgs e)
		{
			NewCommandTab();
		}

		private void OnSettingsMenuItemClicked(object sender, EventArgs e)
		{
			using (var dlg = new SettingsDialog())
			{
				dlg.ApplyTabSettings(_tabSettings);
				if (dlg.ShowDialog(this) == DialogResult.OK)
				{
					var tabSettings = new TabSettings(dlg.TextFont, dlg.PromptColor, dlg.TextColor, dlg.BackgroundColor);
					_applicationSettingsManager.WriteTabSettings(tabSettings);

					foreach (TabPage tab in _commandTabContainer.TabPages)
					{
						_tabSettings = tabSettings;
						((ShellTabPresenter)tab.Tag).ApplySettings(_tabSettings);
					}
				}
			}
		}

		private void OnTabContainerDoubleClicked(object sender, EventArgs e)
		{
			var tab = GetTabFromPoint(_commandTabContainer.PointToClient(MousePosition));
			if (tab != null)
			{
				if (_tabContainersLastMouseAction == MouseButtons.Left)
				{
					RenameTab(tab);
				}
			}
		}

		private TabPage GetTabFromPoint(Point point)
		{
			for (int i = 0; i < _commandTabContainer.TabCount; i++)
			{
				if (_commandTabContainer.GetTabRect(i).Contains(point))
				{
					return _commandTabContainer.TabPages[i];
				}
			}
			
			return null;
		}

		private void OnTabContainerMouseUp(object sender, MouseEventArgs e)
		{
			_tabContainersLastMouseAction = e.Button;
			if (e.Button == MouseButtons.Middle)
			{
				var tab = GetTabFromPoint(e.Location);
				if (tab != null)
				{
					CloseTab(tab);
				}
			}
		}

		private void OnCloseTabItemClicked(object sender, EventArgs e)
		{
			CloseCurrentTab();
		}

		private void OnCloseAllTabsItemClicked(object sender, EventArgs e)
		{
			CloseAllTabs();
		}

		private void OnAboutAppItemClicked(object sender, EventArgs e)
		{
			using (var dlg = new AboutAppDialog())
			{
				dlg.ShowDialog(this);
			}
		}

		private void OnClipboardItemClicked(object sender, EventArgs e)
		{
			if (sender == miEditCopy)
			{
				Copy();
			}
			else if (sender == miEditPaste)
			{
				Paste();
			}
		}

		private void OnExitAppItemClicked(object sender, EventArgs e)
		{
			Close();
		}

		private void OnFullScreenItemClicked(object sender, EventArgs e)
		{
			if (!miViewFullScreen.Checked)
			{
				// Store "normal" values
				_normalBorderStyle = FormBorderStyle;
				_normalLocation = Location;
				_normalSize = Size;

				FormBorderStyle = FormBorderStyle.None;
				
				Screen screen = Screen.FromHandle(Handle);
				Size = screen.Bounds.Size;
				Location = screen.Bounds.Location;
				TopMost = true;
			}
			else
			{
				FormBorderStyle = _normalBorderStyle;
				Location = _normalLocation;
				Size = _normalSize;
				TopMost = false;
			}
			
			miViewFullScreen.Checked = !miViewFullScreen.Checked;
		}

		private void OnShowStatusBarItemClicked(object sender, EventArgs e)
		{
			statusBar.Visible = !statusBar.Visible;
			miViewStatusBar.Checked = statusBar.Visible;
		}

		private void OnShowMenuBarClicked(object sender, EventArgs e)
		{
			mainMenu.Visible = !mainMenu.Visible;
			miViewMenuBar.Checked = mainMenu.Visible;
		}

		private void OnNewTabPowerShellClicked(object sender, EventArgs e)
		{
			NewCommandTab(ShellTypes.PowerShell);
		}

		private void OnClipboardHistoryClicked(object sender, EventArgs e)
		{
			_clipboardHistoryPresenter.Show();
		}

		private void OnLoadedCommandInfoClicked(object sender, EventArgs e)
		{
			using (CommandsInfoDialogPresenter p = new CommandsInfoDialogPresenter(_commandHost))
			{
				p.Init();
			p.Show(this);
			}
		}
	}
}
