﻿using System;
using System.IO;
using System.Linq;
using Commander.API;
using Commander.Core.Command;
using NUnit.Framework;
using Rhino.Mocks;

namespace Commander.Core.Tests.Command
{
	[TestFixture]
	public class CommandManagerTests
	{
		private CommandManager _commandManager;
		private MockRepository _mocks;
		private ICommandAccess _outputWriter;
		private ICommand _executor;

		[SetUp]
		public void SetUp()
		{
			_mocks = new MockRepository();

			_outputWriter = _mocks.DynamicMock<ICommandAccess>();
			_executor = _mocks.DynamicMock<ICommand>();
			_commandManager = new CommandManager(_outputWriter);

			_executor.Expect(x => x.Identifier).Return("test");
		}

		[TearDown]
		public void TearDown()
		{
			
		}


		[Test]
		public void RegisteredHandleIsCalled()
		{
			_executor.Expect(x => x.Execute(_outputWriter, new string[] { })).Return("");
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);
			_commandManager.Execute("test");

			_mocks.Verify(_executor);
		}

		[Test]
		public void RegisteredHandleIsCalledWhenCommandHasArguments()
		{
			_executor.Expect(x => x.Execute(_outputWriter, new[] { "arg0" })).Return("");
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);
			_commandManager.Execute("test -arg0");

			_mocks.Verify(_executor);
		}

		[Test]
		public void RegisteredHandleIsCalledWithAllArgumentsGivenToCommandExcludingCommandName()
		{
			_executor.Expect(x => x.Execute(_outputWriter, new[] { "arg0", "arg1", "arg2" })).Return("");
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);

			_commandManager.Execute("test -arg0 -arg1 -arg2");
			_mocks.Verify(_executor);
		}


		[Test]
		public void CommandResultIsWrittenToOutput()
		{
			const string commandResponse = "This is a command response";
			_executor.Expect(x => x.Execute(_outputWriter, null)).IgnoreArguments().Return(commandResponse);
			_outputWriter.Expect(x => x.AppendOutput(Environment.NewLine + commandResponse, true, true));
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);

			_commandManager.Execute("test");
			_mocks.Verify(_outputWriter);
		}


		[Test]
		public void RegisterCommandOverridesPrevHandle()
		{
			const string output = "Command 2 output";
			var executor2 = _mocks.DynamicMock<ICommand>();
			executor2.Expect(x => x.Identifier).Return("test");
			executor2.Expect(x => x.Execute(_outputWriter, null)).IgnoreArguments().Return(output);
			_outputWriter.Expect(x => x.AppendOutput(Environment.NewLine + output, true, true));
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);
			_commandManager.RegisterHandler(executor2);
			_commandManager.Execute("test");

			_mocks.Verify(executor2);
			_mocks.Verify(_outputWriter);
		}

		[Test]
		public void NoRegisteredHandlerFoundForCommand()
		{
			_mocks.ReplayAll();
			_commandManager.RegisterHandler(_executor);
			Assert.IsFalse(_commandManager.Execute("bad-cmd"));
		}


		[Test]
		public void RegisteredHandleIsCalledWhenCommandHasArgumentsWithSpaces()
		{
			_executor.Expect(x => x.Execute(_outputWriter, new[] { "My Special Argument" })).Return("");
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);
			_commandManager.Execute("test \"My Special Argument\"");

			_mocks.Verify(_executor);
		}

		[Test]
		public void RegisteredHandleIsCalledWhenCommandHasMultipleArgumentsWithSpaces()
		{
			_executor.Expect(x => x.Execute(_outputWriter, new[] { "My Special Argument", "arg2" })).Return("");
			_mocks.ReplayAll();

			_commandManager.RegisterHandler(_executor);
			_commandManager.Execute("test \"My Special Argument\" arg2");

			_mocks.Verify(_executor);
		}

		[Test]
		public void ChangeDriveCommandHandlesValidDrive()
		{
			DriveInfo driveInfo = DriveInfo.GetDrives().Where(x => x.DriveType == DriveType.Fixed).FirstOrDefault();
			_outputWriter.Expect(x => x.SetDirectory(driveInfo.Name));
			_mocks.ReplayAll();

			_commandManager.Execute(driveInfo.RootDirectory.Name);

			_mocks.Verify(_outputWriter);
		}

		[Test]
		public void ChangeDriveCommandHandlesValidDriveWithNoEndingBackslash()
		{
			DriveInfo driveInfo = DriveInfo.GetDrives().Where(x => x.DriveType == DriveType.Fixed).LastOrDefault();
			_outputWriter.Expect(x => x.SetDirectory(driveInfo.RootDirectory.FullName));
			_mocks.ReplayAll();

			string name = driveInfo.Name.Trim('\\');
			_commandManager.Execute(name);

			_mocks.Verify(_outputWriter);
		}
	}
}
