﻿using System.Collections.Generic;
using System.IO;
using System.Threading;
using NUnit.Framework;

namespace Commander.Core.Tests
{
	[TestFixture]
	public class CommandHistoryManagerTests
	{
		private CommandHistoryManager _historyManager;
		private FileInfo _historyFile;

		#region Setup
		[TestFixtureSetUp]
		public void TestFixtureSetUp()
		{
			
		}

		[SetUp]
		public void SetUp()
		{
			_historyFile = new FileInfo(Path.GetTempFileName());
			_historyManager = new CommandHistoryManager(_historyFile);
			_historyManager.Init();
		}

		[TearDown]
		public void TearDown()
		{
			_historyFile.Delete();
		}
		#endregion Setup


		#region Tests
		[Test]
		public void Init()
		{
			WriteCommandsToFile("dir", "pwd");

			ManualResetEvent resetEvent = new ManualResetEvent(false);
			_historyManager.HistoryLoadComplete += (sender, e) => resetEvent.Set();
			_historyManager.Init();

			resetEvent.WaitOne(1000);
			
			Assert.AreEqual(2, _historyManager.Records.Count);
			Assert.AreEqual("dir", _historyManager.Records[0]);
			Assert.AreEqual("pwd", _historyManager.Records[1]);
		}

		private void WriteCommandsToFile(params string[] commands)
		{
			using (StreamWriter wr = new StreamWriter(_historyFile.OpenWrite()))
			{
				foreach (var command in commands)
				{
					wr.WriteLine(command);
				}

				wr.Close();
			}
		}

		private List<string> ReadCommandsFromFile()
		{
			var retVal = new List<string>();
			using (StreamReader sr = new StreamReader(_historyFile.OpenRead()))
			{
				while (!sr.EndOfStream)
				{
					retVal.Add(sr.ReadLine());
				}

				sr.Close();
			}
			
			return retVal;
		}

		[Test]
		public void AddHistoryEntry()
		{
			_historyManager.Add("cd ..");
			CollectionAssert.AreEqual(new List<string> { "cd .." }, 
									  _historyManager.Records);
		}

		[Test]
		public void AddingMultipleHistoryEntriesWillReturnAll()
		{
			_historyManager.Add("cd ..");
			_historyManager.Add("pwd");
			_historyManager.Add("ls");
			CollectionAssert.AreEqual(new List<string> { "cd ..", "pwd", "ls" }, 
									  _historyManager.Records);
		}

		[Test]
		public void AddingHistoryEntryPersistsToHistoryFile()
		{
			_historyManager.Add("pwd");
			_historyManager.Add("dir");
			
			List<string> commands = ReadCommandsFromFile();
			Assert.AreEqual(2, commands.Count);
			Assert.AreEqual("pwd", commands[0]);
			Assert.AreEqual("dir", commands[1]);
		}

		[Test]
		public void GetMatchingReturnsMatch()
		{
			_historyManager.Add("pwd");
			CollectionAssert.AreEqual(new List<string> { "pwd" },
									  _historyManager.GetMatching("pwd"));
		}

		[Test]
		public void GetMatchingReturnsOnlyMatchesMatchingBeginning()
		{
			_historyManager.Add("pwd");
			_historyManager.Add("cls");
			_historyManager.Add("dir");
			_historyManager.Add("pwd2");
			CollectionAssert.AreEqual(new List<string> { "pwd", "pwd2" },
									  _historyManager.GetMatching("pwd"));
		}

		[Test]
		public void GetMatchingReturnsOnlyMatchesMatchingHonorsCasing()
		{
			_historyManager.Add("pwd");
			_historyManager.Add("cls");
			_historyManager.Add("dir");
			_historyManager.Add("pWd");
			CollectionAssert.AreEqual(new List<string> { "pwd" },
									  _historyManager.GetMatching("pw"));
		}

		[Test]
		public void ClearingHistoryRemovesCachedListAndDeletesFile()
		{
			_historyManager.Add("pwd");
			_historyManager.Add("cls");
			_historyManager.Add("dir");
			_historyManager.Add("pWd");

			_historyManager.Clear();

			_historyFile.Refresh();
			Assert.AreEqual(0, _historyManager.Records.Count);
			Assert.IsFalse(_historyFile.Exists);
		}
		#endregion Tests
	}
}
