﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using Commander.API;

namespace Commander.Core
{
	public class CommandHistoryManager : ICommandHistory
	{
		private readonly FileInfo _historyFile;
		private readonly List<string> _commandHistory;

		public event EventHandler<EventArgs> HistoryLoadComplete;

		public CommandHistoryManager(FileInfo historyFile)
		{
			_historyFile = historyFile;
			_commandHistory = new List<string>(50);
		}

		public void Init()
		{
			Trace.WriteLine("Trying to load history from: " + _historyFile.FullName);

			_historyFile.Refresh();
			Thread t = new Thread(LoadHistoryFromFile) {
			    IsBackground = true,
			    Priority = ThreadPriority.Lowest,
			    Name = "Command.History.FileLoader"
			};
			t.Start();
		}

		private void LoadHistoryFromFile()
		{
			var file = _historyFile.Open(FileMode.OpenOrCreate);
			using (var sr = new StreamReader(file, Encoding.UTF8))
			{
				while (!sr.EndOfStream)
				{
					var line = sr.ReadLine();
					if (String.IsNullOrEmpty(line))
					{
						continue;
					}

					_commandHistory.Add(line);
				}
			}

			file.Close();
			Trace.WriteLine("Loaded " + _commandHistory.Count + " lines from history...");
			if (HistoryLoadComplete != null)
			{
				HistoryLoadComplete.Invoke(this, EventArgs.Empty);
			}
		}

		public void Add(string command)
		{
			_commandHistory.Add(command);
			WriteToFile(command);
		}

		public void Clear()
		{
			_commandHistory.Clear();
			
			_historyFile.Delete();
			_historyFile.Refresh();
		}

		public List<string> GetMatching(string command)
		{
			return _commandHistory.Where(x => x.StartsWith(command, StringComparison.CurrentCulture)).ToList();
		}

		public List<string> Records
		{
			get { return _commandHistory; }
		}

		private void WriteToFile(string command)
		{
			using (var file = _historyFile.Open(FileMode.Append, FileAccess.Write, FileShare.Read))
			{
				byte[] bytes = Encoding.UTF8.GetBytes(command + Environment.NewLine);
				file.Write(bytes, 0, bytes.Length);

				file.Close();
			}
		}


		public void Dispose()
		{
			
		}
	}
}
