﻿using System;
using System.Drawing;
using System.Windows.Forms;
using Commander.Core;

namespace Commander.UI.Controls.Text
{
	public class CommanderTextBox : RichTextBox, IClipboardActions
	{
		private readonly string _prompt;

		/// <summary>
		/// Initializes a new instance of the <see cref="CommanderTextBox"/> class.
		/// </summary>
		/// <param name="prompt">The prompt.</param>
		public CommanderTextBox(string prompt)
		{
			_prompt												= prompt;
			WordWrap											= false;
			AcceptsTab											= true;
			ScrollBars											= RichTextBoxScrollBars.ForcedVertical;
			RichTextShortcutsEnabled							= false;
		}

		#region Public Methods
		/// <summary>
		/// Returns the user written input string that's on the current line.
		/// </summary>
		/// <param name="actualStartIndex">Actual index of the start.</param>
		/// <returns></returns>
		public string GetCurrentLineString(int actualStartIndex = 0)
		{
			int currentLinePosition = GetCurrentLineStartPosition();
			if (actualStartIndex > 0)
			{
				currentLinePosition = actualStartIndex;
			}

			if (currentLinePosition > TextLength)
			{
				// Will make it a empty string...
				currentLinePosition = TextLength;
			}

			return Text.Substring(currentLinePosition, TextLength - currentLinePosition).Trim();
		}

		public void ReplaceCurrentLineWith(string newText)
		{
			var startPosition = GetCurrentLineStartPosition();
			Select(startPosition, TextLength - startPosition);
			SelectedText = newText;
			SelectionLength = 0;
			SelectionStart = TextLength;
		}

		/// <summary>
		/// Scrolls to bottom of text box.
		/// </summary>
		public void ScrollToBottom()
		{
			CommanderWin32.SendMessage(Handle, CommanderWin32.WM_VSCROLL, CommanderWin32.SB_BOTTOM, 0);
		}

		/// <summary>
		/// Clears the output.
		/// </summary>
		public void ClearOutput()
		{
			ResetText();
			Clear();

			AppendOutput("", false, true, null, false);
			UpdatePosition();
		}

		/// <summary>
		/// Appends the output to the text box applying color etc according to the call.
		/// </summary>
		/// <param name="output">The output.</param>
		/// <param name="newLine">if set to <c>true</c> [new line].</param>
		/// <param name="addPrompt"></param>
		/// <param name="textColor">Color of the text.</param>
		/// <param name="skipStyle">if set to <c>true</c> [skip style].</param>
		public void AppendOutput(string output, bool newLine = false, bool addPrompt = false, Color? textColor = null, bool skipStyle = true)
		{
			SuspendLayout();

			SelectionStart = TextLength;
			SelectionLength = 0;
			int start = TextLength;
				AppendText(output);
			int end = TextLength;

			if (!skipStyle)
			{
				Select(start, end - start);
				SelectionColor = (textColor != null ? textColor.Value : ForeColor);
			}

			if (newLine)
			{
				AppendText(Environment.NewLine);
			}

			if (addPrompt)
			{
				Color last = SelectionColor;
				start = TextLength;
				AppendText(_prompt);
				end = TextLength;

				Select(start, end - start);
				SelectionColor = PromptColor;

				Select(end - 1, 1);
				SelectionColor = last;
				SelectionLength = 0;
			}

			var selectEnd = (end - start);
			if (selectEnd > 0)
			{
				selectEnd -= 1;
			}

			if (!skipStyle)
			{
				SelectionStart = selectEnd;
				SelectionLength = 1;
				SelectionColor = Color.FromArgb(255, ForeColor);
				SelectionFont = new Font(Font, FontStyle.Regular);
			}

			ClearUndo();
			ResumeLayout();
			UpdatePosition();
		}

		/// <summary>
		/// Updates the position of the carred and makes sure that the text box is scrolled to the bottom.
		/// </summary>
		public void UpdatePosition()
		{
			SelectionStart = TextLength;
			SelectionLength = 0;

			ScrollToBottom();
		}

		/// <summary>
		/// Returns the current line the carret is on
		/// </summary>
		/// <returns></returns>
		public int GetCurrentLine()
		{
			return GetLineFromCharIndex(SelectionStart);
		}

		#endregion


		#region IClipboardActions
		/// <summary>
		/// Copies the current selection in the text box to the Clipboard.
		/// </summary>
		/// <PermissionSet>
		///   <IPermission class="System.Security.Permissions.EnvironmentPermission, mscorlib, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Unrestricted="true"/>
		///   <IPermission class="System.Security.Permissions.FileIOPermission, mscorlib, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Unrestricted="true"/>
		///   <IPermission class="System.Security.Permissions.SecurityPermission, mscorlib, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Flags="UnmanagedCode, ControlEvidence"/>
		///   <IPermission class="System.Diagnostics.PerformanceCounterPermission, System, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Unrestricted="true"/>
		///   </PermissionSet>
		public new virtual void Copy()
		{
			var text = SelectedText;
			if (!String.IsNullOrEmpty(text))
			{
				Clipboard.SetText(text, TextDataFormat.Text);
			}
		}

		/// <summary>
		/// Replaces the current selection in the text box with the contents of the Clipboard.
		/// </summary>
		/// <PermissionSet>
		///   <IPermission class="System.Security.Permissions.EnvironmentPermission, mscorlib, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Unrestricted="true"/>
		///   <IPermission class="System.Security.Permissions.FileIOPermission, mscorlib, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Unrestricted="true"/>
		///   <IPermission class="System.Security.Permissions.SecurityPermission, mscorlib, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Flags="UnmanagedCode, ControlEvidence"/>
		///   <IPermission class="System.Diagnostics.PerformanceCounterPermission, System, Version=2.0.3600.0, Culture=neutral, PublicKeyToken=b77a5c561934e089" version="1" Unrestricted="true"/>
		///   </PermissionSet>
		public new virtual void Paste()
		{
			string text = Clipboard.GetText();
			if (!String.IsNullOrEmpty(text))
			{
				AppendOutput(text);
			}
		}

		public void Paste(string content)
		{
			AppendOutput(content);
		}
		#endregion


		#region Protected Methods
		/// <summary>
		/// Places the carret on last line.
		/// </summary>
		protected void PlaceCarretOnLastLine()
		{
			int currentLine = GetCurrentLine();
			if (currentLine != Lines.Length - 1)
			{
				SelectionStart = TextLength;
				SelectionLength = 0;
			}

			int lineStartPosition = GetCurrentLineStartPosition();
			if (SelectionStart < lineStartPosition)
			{
				SelectionStart = lineStartPosition;
				SelectionLength = 0;
			}
		}

		/// <summary>
		/// Get's the starting char index for a line based on if a prompt is present or not.
		/// </summary>
		/// <returns></returns>
		protected int GetCurrentLineStartPosition()
		{
			int currentLine = GetCurrentLine();
			int availableLines = Lines.Length - 1;
			if (currentLine >= availableLines)
			{
				currentLine = availableLines;
			}

			return (TextLength - Lines[currentLine].Length) + _prompt.Length;
		}

		/// <summary>
		/// Moves the carret to the last line and the last written character + 1
		/// </summary>
		protected void GoToLineEnd()
		{
			SelectionStart = TextLength;
		}

		/// <summary>
		/// Moves the carret to the last line and the first written character after the prompt - 1
		/// </summary>
		protected void GoToLineStart()
		{
			SelectionStart = GetCurrentLineStartPosition();
		}

		#endregion


		#region Basic Key handling
		/// <summary>
		/// Raises the <see cref="E:System.Windows.Forms.Control.KeyDown"/> event.
		/// </summary>
		/// <param name="e">A <see cref="T:System.Windows.Forms.KeyEventArgs"/> that contains the event data.</param>
		protected override void OnKeyDown(KeyEventArgs e)
		{
			int carretCurrentLineStart = GetCurrentLineStartPosition();
			int carretCurrentPosition = SelectionStart;

			if (e.KeyCode != Keys.ControlKey)
			{
				PlaceCarretOnLastLine();
			}

			// Hinder stepping left off the prompt
			HandleKeyNavigationLeft(e, carretCurrentLineStart);

			if (e.KeyCode == Keys.Up || e.KeyCode == Keys.Down || e.KeyCode == Keys.Home || e.KeyCode == Keys.End)
			{
				if (e.KeyCode == Keys.Up || e.KeyCode == Keys.Home)
				{
					GoToLineStart();
				}
				else
				{
					GoToLineEnd();
				}

				if (e.Shift)
				{
					// Current line selection (Prompt selection)
					if (e.KeyCode == Keys.End)
					{
						Select(carretCurrentPosition, TextLength - carretCurrentPosition);
					}
					else if (e.KeyCode == Keys.Home)
					{
						Select(carretCurrentLineStart, carretCurrentPosition - carretCurrentLineStart);
					}
				}

				e.Handled = true;
				e.SuppressKeyPress = true;
			}

			if (e.KeyCode == Keys.Back)
			{
				if (!String.IsNullOrEmpty(SelectedText))
				{
					SelectedText = "";
					e.Handled = true;
					e.SuppressKeyPress = true;
				}
			}

			if (e.Control)
			{
				HandleControlKeyActions(e, carretCurrentLineStart);
			}

			base.OnKeyDown(e);
		}

		/// <summary>
		/// Handles the key navigation going left.
		/// </summary>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyEventArgs"/> instance containing the event data.</param>
		/// <param name="caretPosition">The caret position.</param>
		private void HandleKeyNavigationLeft(KeyEventArgs e, int caretPosition)
		{
			if (e.KeyCode == Keys.Back || e.KeyCode == Keys.Left)
			{
				if (SelectionStart <= caretPosition)
				{
					SelectionStart = caretPosition;
					e.Handled = true;
					e.SuppressKeyPress = true;
				}
			}
		}

		/// <summary>
		/// Handles the control key actions.
		/// </summary>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyEventArgs"/> instance containing the event data.</param>
		/// <param name="caretPosition">The caret position.</param>
		private void HandleControlKeyActions(KeyEventArgs e, int caretPosition)
		{
			if (e.KeyCode == Keys.L)
			{
				ClearOutput();
			}

			if (e.KeyCode == Keys.A)
			{
				Select(caretPosition, 0);
				e.SuppressKeyPress = true;
				e.Handled = true;
			}

			if (e.KeyCode == Keys.E || e.KeyCode == Keys.R)
			{
				// Disable alignement
				e.Handled = true;
				e.SuppressKeyPress = true;
			}
		}

		#endregion


		#region Properties
		public Color PromptColor { get; set; }
		#endregion
	}
}
