﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using Commander.API;
using Commander.UI.Views;

namespace Commander.UI.Presenters
{
	public class AutoCompletePresenter : IPresenter<AutoCompleteView>
	{
		private readonly AutoCompleteView _view;
		private readonly Control _attachedTo;
		private readonly ICommandHistory _commandHistory;
		public event EventHandler<AutoCompleteItemClickedArgs> ItemClicked;

		public AutoCompletePresenter(Control attachedTo, ICommandHistory commandHistory)
			: this(new AutoCompleteView(), attachedTo, commandHistory)
		{ }

		public AutoCompletePresenter(AutoCompleteView view, Control attachedTo, ICommandHistory commandHistory)
		{
			_view = view;
			_attachedTo = attachedTo;
			_commandHistory = commandHistory;
		}

		public void Init()
		{
			_view.Parent = _attachedTo;
			
			// Redirect focus back to the attached owning control if we would receive it.
			_view.GotFocus += (sender, e) => _view.ItemsList.Focus();
			_view.ItemsList.DoubleClick += (sender, e) => TriggerItemClickedEvent();
			_view.ClearHistoryButton.Click += (sender, e) => ClearHistory();
			_view.Visible = false;
		}

		private void ClearHistory()
		{
			_commandHistory.Clear();
			ShowItems();
		}

		private void TriggerItemClickedEvent()
		{
			if (_view.ItemsList.SelectedIndex < 0)
			{
				return;
			}

			if (ItemClicked != null)
			{
				string command = (string)_view.ItemsList.Items[_view.ItemsList.SelectedIndex];
				ItemClicked.Invoke(this, new AutoCompleteItemClickedArgs(command));
			}
		}

		public void NavigateUp()
		{
			if (_view.ItemsList.Items.Count < 1)
			{
				return;
			}

			int current = _view.ItemsList.SelectedIndex - 1;
			if (current < 0)
			{
				current = _view.ItemsList.Items.Count - 1;
			}

			_view.ItemsList.SelectedIndex = current;
		}

		public void NavigateDown()
		{
			if (_view.ItemsList.Items.Count < 1)
			{
				return;
			}

			int current = _view.ItemsList.SelectedIndex + 1;
			if (current >= _view.ItemsList.Items.Count)
			{
				current = 0;
			}

			_view.ItemsList.SelectedIndex = current;
		}

		public void Show()
		{
			Show(_attachedTo.FindForm(), new Point(15, 60));
		}

		public void Show(Control attachTo, Point point)
		{
			_view.Parent = attachTo.FindForm();
			_view.Show();
			point = new Point(point.X + 15, point.Y + 60);
			var endPoint = new Point(point.X + _view.Width, point.Y + _view.Height);

			if (_attachedTo.ClientRectangle.Bottom < endPoint.Y)
			{
				point.Offset(0, -((endPoint.Y - _attachedTo.ClientRectangle.Bottom) + 20));
			}

			if (_attachedTo.ClientRectangle.Right < endPoint.X)
			{
				point.Offset(-(endPoint.X - _attachedTo.ClientRectangle.Right), 0);
			}

			_view.ForeColor = Color.Black;
			_view.BackColor = Color.White;
			_view.Location = point;
			_view.BringToFront();
			_view.Show();
			ShowItems();
		}

		public void Hide()
		{
			_view.Hide();
		}


		public void ShowItems()
		{
			_view.ItemsList.SuspendLayout();
			_view.ItemsList.Items.Clear();
			
			List<string> records = _commandHistory.Records;
			records.Reverse();
			_view.ItemsList.Items.AddRange(records.ToArray());
			_view.BottomCountLabel.Text = String.Format("{0} items", records.Count);
			_view.BottomCountLabel.ForeColor = _view.ForeColor;
			_view.ItemsList.ResumeLayout(true);
		}

		public string GetSelectedCommand()
		{
			List<string> records = _commandHistory.Records;
			if (_view.ItemsList.SelectedIndex < 0 || _view.ItemsList.SelectedIndex >= records.Count - 1)
			{
				return null;
			}

			return records[_view.ItemsList.SelectedIndex];
		}

		#region Properties
		public bool IsVisible
		{
			get { return _view.Visible; }
		}

		public AutoCompleteView View
		{
			get { return _view; }
		}

		#endregion

		public void SelectCurrentItem()
		{
			TriggerItemClickedEvent();
		}
	}
}
