﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using Commander.Core;
using Commander.UI.Views;

namespace Commander.UI.Presenters
{
	public class ClipboardHistoryPresenter : IPresenter<ClipboardHistoryView>, IDisposable
	{
		private readonly ClipboardHistoryView _view;
		private readonly IClipboardHistory _clipboardHistory;
		private readonly IClipboardActions _output;
		private readonly Form _owner;
		private readonly CommanderDialog _form;

		public ClipboardHistoryPresenter(IClipboardHistory clipboardHistory, IClipboardActions output, Form owner)
			: this(new ClipboardHistoryView(), clipboardHistory, output, owner)
		{ }

		public ClipboardHistoryPresenter(ClipboardHistoryView view, IClipboardHistory clipboardHistory, IClipboardActions output, Form owner)
		{
			_view = view;
			_clipboardHistory = clipboardHistory;
			_output = output;
			_owner = owner;
			_form = new CommanderDialog();
		}

		public void Init()
		{
			_view.Dock = DockStyle.Fill;
			_form.Controls.Add(_view);
			_form.ShowIcon = false;
			_form.FormBorderStyle = FormBorderStyle.Sizable;
			_form.Size = new Size(450, 480);
			_form.TopMost = true;
			_form.StartPosition = FormStartPosition.CenterParent;
			_form.MinimizeBox = false;
			_form.MaximizeBox = false;
			_form.Text = "Clipboard History";
			_form.AcceptButton = _view.OKButton;
			_form.CancelButton = _view.CancelButton;
			_form.Shown += OnFormShown;

			_view.CancelButton.Click += (sender, e) => { _form.DialogResult = DialogResult.Cancel; _form.Close(); };
			_view.OKButton.Click += (sender, e) => { _form.DialogResult = DialogResult.OK; _form.Close(); };
			_view.ClipboardHistoryList.SelectedIndexChanged += OnClipboardHistoryListIndexChanged;
		}

		public void Show()
		{
			ShowItems();

			if (_form.ShowDialog(_owner) == DialogResult.OK)
			{
				if (_view.ClipboardHistoryList.SelectedIndex < 0)
				{
					return;
				}

				string item = (string)_view.ClipboardHistoryList.Items[_view.ClipboardHistoryList.SelectedIndex];
				if (!String.IsNullOrEmpty(item))
				{
					Clipboard.SetText(item);
					_output.Paste(item);
				}
			}
		}

		public void ShowItems()
		{
			_view.ClipboardHistoryList.Items.Clear();
			List<string> records = _clipboardHistory.Records;
			records.ForEach(r => _view.ClipboardHistoryList.Items.Add(r));

			string current = _clipboardHistory.CurrentClipboardContent.Trim();
			if (!String.IsNullOrEmpty(current))
			{
				_view.ClipboardHistoryList.Items.Insert(0, current);
			}
		}

		public void Hide()
		{
			_form.Hide();
		}

		public bool IsVisible
		{
			get { return _view.Visible; }
		}

		public ClipboardHistoryView View
		{
			get { return _view; }
		}

		public void Dispose()
		{
			if (_view != null)
			{
				_view.Dispose();
			}

			if (_form != null)
			{
				_form.Dispose();
			}
		}

		private void OnFormShown(object sender, EventArgs e)
		{
			if (_view.ClipboardHistoryList.Items.Count > 0)
			{
				_view.ClipboardHistoryList.SelectedIndex = 0;
			}

			_view.ClipboardHistoryList.Focus();
		}

		private void OnClipboardHistoryListIndexChanged(object sender, EventArgs e)
		{
			if (_view.ClipboardHistoryList.SelectedIndex < 0)
			{
				return;
			}

			_view.SelectedText.Text = (string)_view.ClipboardHistoryList.Items[_view.ClipboardHistoryList.SelectedIndex];
		}
	}
}
