﻿using System;
using System.Linq;
using System.Reflection;
using System.Windows.Forms;
using Commander.API;
using Commander.Core;
using Commander.Core.Command;
using Commander.UI.Views;

namespace Commander.UI.Presenters
{
	public class CommandsInfoDialogPresenter : IPresenter<CommandsInfoDialog>, IDisposable
	{
		private readonly CommandHost _commandHost;

		public CommandsInfoDialogPresenter(CommandHost commandHost)
		{
			_commandHost = commandHost;
		}


		private void SyncCommandsView()
		{
			View.CommandsList.SuspendLayout();

			View.CommandsList.Items.Clear();
			ClearAllUiElements();
			IOrderedEnumerable<ICommand> commands = _commandHost.LoadedCommands.OrderBy(c => c.Identifier);
			foreach (ICommand command in commands)
			{
				ListViewItem item = new ListViewItem(command.Identifier) {
					Tag = command
				};

				item.SubItems.Add(command.GetType().Assembly.GetName().Version.ToString());
				View.CommandsList.Items.Add(item);
			}

			View.CommandsList.ResumeLayout(true);
			if (View.CommandsList.Items.Count > 0)
			{
				View.CommandsList.Items[0].Selected = true;
			}
		}

		private void OnCommandItemChanged(object sender, EventArgs e)
		{
			ClearAllUiElements();
			if (View.CommandsList.SelectedItems.Count < 1)
			{
				return;
			}

			ICommand command = View.CommandsList.SelectedItems[0].Tag as ICommand;
			if (command == null)
			{
				return;
			}

			CommandToUi(command);
		}

		private void CommandToUi(ICommand command)
		{
			Assembly assembly = command.GetType().Assembly;

			View.VersionLabel.Text = assembly.GetVersion();
			View.CompanyLabel.Text = "Author: " + assembly.GetCompanyName();
			View.TitleLabel.Text = assembly.GetTitle();
			View.CopyrightLabel.Text = assembly.GetCopyright();
			View.DescriptionLabel.Text = assembly.GetDescription();
		}

		private void ClearAllUiElements()
		{
			View.VersionLabel.Text = "";
			View.CompanyLabel.Text = "";
			View.TitleLabel.Text = "";
			View.CopyrightLabel.Text = "";
			View.DescriptionLabel.Text = "";
		}

		#region IPresenter
		public void Init()
		{ }

		public void Show()
		{
			Show(null);
		}

		public void Show(Form owner)
		{
			if (View == null)
			{
				View = new CommandsInfoDialog();
				View.CloseButton.Click += (sender, e) => View.Close();
				View.CommandsList.SelectedIndexChanged += OnCommandItemChanged;
			}

			if (owner != null)
			{
				View.Icon = owner.Icon;
			}

			SyncCommandsView();

			View.ShowInTaskbar = false;
			View.ShowDialog(owner);
		}

		public void Hide()
		{
			if (IsVisible)
			{
				View.Close();
			}
		}

		public bool IsVisible
		{
			get { return View != null && View.Visible; }
		}

		public CommandsInfoDialog View { get; private set; }
		#endregion

		public void Dispose()
		{
			if (View != null && !View.IsDisposed)
			{
				View.Dispose();
				View = null;
			}
		}
	}
}
