﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Commander.API;
using Commander.Core;
using Commander.Core.Command;
using Commander.UI.Controls.Text;
using Microsoft.WindowsAPICodePack.Taskbar;

namespace Commander.UI.Presenters
{
	public class ShellTabPresenter : IPresenter<TabPage>, ICommandAccess, IClipboardActions
	{
		private readonly string _prompt;
		private readonly ICommandTabHost _tabHost;
		private readonly ICommandManager _internalCommandManager;
		private readonly ICommandHistory _commandHistoryManager;
		private readonly IClipboardHistory _clipboardHistory;
		private readonly AutoCompletePresenter _autoComplete;
		private readonly TabPage _tabPage;
		private TabSettings _tabSettings;
		private readonly ShellTypes _type;
		private readonly CommanderTextBox _commandWindow;
		private string _workingDirectory;
		private readonly Encoding _targetEncoding;
		private Thread _commandThread;
		private StreamWriter _activeStdIn;
		private Process _activeProcesss;
		private int _lastOutputIndex;
		private readonly CommandWindowPopupMenuPresenter _contextMenuPresenter;
		private readonly CommanderDragDropPresenter _dragDropPresenter;


		public ShellTabPresenter(ICommandTabHost tabHost, ICommandManager internalCommandManager, 
								 ICommandHistory commandHistoryManager, IClipboardHistory clipboardHistory, AutoCompletePresenter autoComplete,
								 Encoding encoding, string prompt, TabSettings tabSettings, ShellTypes type)
			: this(tabHost, internalCommandManager, commandHistoryManager, clipboardHistory, autoComplete, encoding, prompt, new CommanderTextBox(prompt), new TabPage(prompt), tabSettings, type)
		{ }

		public ShellTabPresenter(ICommandTabHost tabHost, ICommandManager internalCommandManager, 
								 ICommandHistory commandHistoryManager, IClipboardHistory clipboardHistory, AutoCompletePresenter autoComplete, 
								 Encoding encoding, string prompt, CommanderTextBox commandWindow,
								 TabPage tabPage, TabSettings tabSettings, ShellTypes type)
		{
			_tabHost = tabHost;
			_internalCommandManager = internalCommandManager;
			_commandHistoryManager = commandHistoryManager;
			_clipboardHistory = clipboardHistory;
			_autoComplete = autoComplete;
			_targetEncoding = encoding;
			_prompt = prompt;
			_commandWindow = commandWindow;
			_tabPage = tabPage;
			_tabSettings = tabSettings;
			_type = type;
			_contextMenuPresenter = new CommandWindowPopupMenuPresenter(_commandWindow, this, _clipboardHistory);
			_dragDropPresenter = new CommanderDragDropPresenter(_commandWindow, this);
		}

		public void Init()
		{
			_workingDirectory = Environment.GetFolderPath(Environment.SpecialFolder.UserProfile);

			_tabPage.Text = _prompt;
			_commandWindow.Parent = _tabPage;
			_commandWindow.Dock = DockStyle.Fill;
			_commandWindow.BorderStyle = BorderStyle.None;
			_commandWindow.Parent = _tabPage;
			_commandWindow.BackColor = Color.Black;
			_commandWindow.ForeColor = Color.White;
			_commandWindow.Font = new Font("Consolas", 11.5F, FontStyle.Regular, GraphicsUnit.Point);
			_commandWindow.WordWrap = false;
			_commandWindow.AcceptsTab = true;
			_commandWindow.ScrollBars = RichTextBoxScrollBars.ForcedVertical;
			_commandWindow.RichTextShortcutsEnabled = false;
			_commandWindow.KeyUp += OnCommandKeyUp;
			_commandWindow.KeyDown += OnCommandKeyDown;
			_commandWindow.LinkClicked += OnCommandWindowLinkClicked;
			
			_tabPage.Tag = this;
			_tabHost.AddTab(_tabPage);
			_commandWindow.Focus();

			ApplySettings(_tabSettings);
			AppendOutput("", false, true);

			_contextMenuPresenter.Init();
			_dragDropPresenter.Init();
		}

		public void Show()
		{
			_tabPage.Visible = true;
			_commandWindow.Focus();
		}

		public void Hide()
		{
			_tabPage.Visible = false;
		}

		public bool IsVisible
		{
			get { return (_commandWindow != null && _commandWindow.Visible); }
		}

		public TabPage View
		{
			get { return _tabPage; }
		}

		public CommanderTextBox CommandWindow
		{
			get { return _commandWindow; }
		}

		private void HandleEnterKey()
		{
			if (_activeStdIn != null)
			{
				_commandWindow.AppendText(Environment.NewLine);
				_activeStdIn.WriteLine(GetCommandWithArgs(_lastOutputIndex));
				return;
			}

			ParseAndExecute();
			InvalidateNeededItems();
		}

		private void KillCurrentJob()
		{
			if (_commandThread != null && _commandThread.IsAlive)
			{
				if (!_activeProcesss.HasExited)
				{
					_activeProcesss.Kill();
				}

				_commandThread.Abort();
			}
		}

		private void ParseAndExecute()
		{
			try
			{
				string cmd = GetCommandWithArgs();
				if (!String.IsNullOrEmpty(cmd))
				{
					// Add to history
					_commandHistoryManager.Add(cmd);

					if (!_internalCommandManager.Execute(cmd))
					{
						Trace.WriteLine("Sending command to be executed by cmd: '" + cmd + "'");
						ExecuteCommand(cmd);
					}

					return;
				}

				// No command, just make a new line.
				AppendOutput("", true, true);
			}
			catch
			{
				AppendOutput("Bad command. I can't understand...", true, true, Color.Red);
			}
		}

		private string GetCommandWithArgs(int correctedStartIndex = 0)
		{
			return _commandWindow.GetCurrentLineString(correctedStartIndex);
		}

		#region ICommandAccess Implementation
		public void AppendOutput(string output, bool newLine, bool addPrompt)
		{
			AppendOutput(output, newLine, addPrompt, null);
		}

		public void ClearOutput()
		{
			_commandWindow.ClearOutput();
		}

		public void SetDirectory(string dir)
		{
			_workingDirectory = dir;
		}

		public string WorkingDirectory
		{
			get { return _workingDirectory; }
		}

		public ICommandHistory History
		{
			get { return _commandHistoryManager; }
		}

		public void CloseInstance()
		{
			_tabHost.CloseTab(_tabPage);
		}
		#endregion

		private void AppendOutput(string output, bool newLine = false, bool addPrompt = false, Color? textColor = null, bool skipStyle = false)
		{
			_commandWindow.AppendOutput(output, newLine, addPrompt, textColor, skipStyle);
			_lastOutputIndex = _commandWindow.TextLength;
			InvalidateNeededItems();
		}

		private void InvalidateNeededItems()
		{
			if (!TaskbarManager.IsPlatformSupported)
			{
				return;
			}

			TabbedThumbnail thumbnailPreview = TaskbarManager.Instance.TabbedThumbnail.GetThumbnailPreview(_commandWindow);
			if (thumbnailPreview != null)
			{
				thumbnailPreview.InvalidatePreview();
			}
		}

		private void ExecuteCommand(string command)
		{
			_commandThread = new Thread(ThreadedProcessExecute) {
				Priority = ThreadPriority.BelowNormal,
				Name = "Command.Executor",
				IsBackground = true
			};

			_commandWindow.AppendText(Environment.NewLine);
			_commandThread.Start(command);
		}


		private void ThreadedProcessExecute(object rawCommand)
		{
			var cmd = (string)rawCommand;
			var args = String.Format("/c {0}", cmd);
			var filename = "cmd.exe";

			if (_type == ShellTypes.PowerShell)
			{
				args = "-NoLogo -NonInteractive -WindowStyle Hidden -Command \"& {" + cmd + "}\"";
				filename = "powershell.exe";
			}

			Debug.WriteLine(String.Format("Executing command {0}, setting working dir to {1}",
										  cmd, _workingDirectory));

			// Start the child process.
			_activeProcesss = new Process {
				StartInfo = {
					WindowStyle = ProcessWindowStyle.Hidden,
					CreateNoWindow = true,
					UseShellExecute = false,
					RedirectStandardOutput = true,
					RedirectStandardError = true,
					RedirectStandardInput = true,
					FileName = filename,
					Arguments = args,
					WorkingDirectory = _workingDirectory,
					StandardOutputEncoding = _targetEncoding,
					StandardErrorEncoding = _targetEncoding
				}
			};

			try
			{
				// Redirect the output stream of the child process.
				_activeProcesss.Start();

				Thread stdErrThread = new Thread(() => ReadStream(_activeProcesss.StandardError, true));
				stdErrThread.IsBackground = true;
				stdErrThread.Priority = ThreadPriority.Lowest;
				stdErrThread.Start();

				_activeStdIn = _activeProcesss.StandardInput;
				_activeStdIn.AutoFlush = true;

				while (true)
				{
					char[] buffer = new char[4096];
					int read = _activeProcesss.StandardOutput.Read(buffer, 0, buffer.Length);

					char[] data = new char[read];
					Array.ConstrainedCopy(buffer, 0, data, 0, read);

					string result = new string(data);
					bool isDone = false;
					if (result.EndsWith(Environment.NewLine + Environment.NewLine + Environment.NewLine))
					{
						isDone = true;
						result = result.TrimEnd(Environment.NewLine.ToCharArray());
					}

					InvokeAppendCommandOutput(result);

					if (isDone || _activeProcesss.StandardOutput.EndOfStream)
					{
						break;
					}
				}

				_activeProcesss.WaitForExit();
			}
			catch (ThreadAbortException abortException)
			{
				Trace.WriteLine("Aborting command " + cmd);
				Trace.WriteLine(abortException);

				InvokeAppendCommandOutput("Aborted command " + cmd, true);
			}
			catch (Exception e)
			{
				Trace.WriteLine("Error: ");
				Trace.WriteLine(e);
			}
			finally
			{
				_activeStdIn = null;
				InvokeAppendNewLine();
				_activeProcesss.Dispose();
			}
		}

		private void ReadStream(StreamReader stream, bool isErrorStream = false)
		{
			string line = stream.ReadLine();
			if (!String.IsNullOrEmpty(line))
			{
				InvokeAppendCommandOutput(line, isErrorStream);
			}
		}

		private void InvokeAppendCommandOutput(string output, bool addNewLine = false, bool isError = false)
		{
			if (_commandWindow.InvokeRequired)
			{
				_tabPage.Invoke(new Action<string, bool, bool>(InvokeAppendCommandOutput), output, addNewLine, isError);
				return;
			}

			AppendOutput(output, false, false, (isError ? Color.Red : new Color?()), (!isError));
		}

		private void InvokeAppendNewLine()
		{
			if (_commandWindow.InvokeRequired)
			{
				_tabPage.Invoke(new Action(InvokeAppendNewLine));
				return;
			}

			AppendOutput("", true, true);
		}

		public void ApplySettings(TabSettings tabSettings)
		{
			_tabSettings = tabSettings;

			_commandWindow.BackColor = _tabSettings.BackgroundColor;
			_commandWindow.ForeColor = _tabSettings.TextColor;
			_commandWindow.PromptColor = _tabSettings.PromptColor;
			_commandWindow.Font = _tabSettings.Font;
		}

		public void Copy()
		{
			_commandWindow.Copy();
		}

		public void Paste()
		{
			_commandWindow.Paste();
		}

		public void Paste(string content)
		{
			_commandWindow.Paste(content);
		}

		private static void OnCommandWindowLinkClicked(object sender, LinkClickedEventArgs e)
		{
			try
			{
				Process.Start(e.LinkText);
			}
			catch (Exception error)
			{
				Trace.WriteLine("Unable to execute link: " + e.LinkText);
				Trace.WriteLine(error);
			}
		}

		private void OnCommandKeyUp(object sender, KeyEventArgs e)
		{
			if (e.Control)
			{
				if (e.KeyCode == Keys.L)
				{
					// We don't allow to clear the screen when in interactive input mode...
					if (_activeStdIn == null)
					{
						_commandWindow.ClearOutput();
					}
				}

				if (e.KeyCode == Keys.X)
				{
					KillCurrentJob();
				}
			}
		}

		private void ShowAutoComplete()
		{
			Point pt = _commandWindow.GetPositionFromCharIndex(_commandWindow.TextLength);
			_autoComplete.Show(_commandWindow, pt);
		}

		private void OnCommandKeyDown(object sender, KeyEventArgs e)
		{
			if (_autoComplete.IsVisible)
			{
				// Update auto complete view's position according to input
				if (e.KeyCode == Keys.Escape)
				{
					_autoComplete.Hide();

					e.Handled = true;
					e.SuppressKeyPress = true;
					return;
				}

				bool handled = false;
				if (e.KeyCode == Keys.Up)
				{
					_autoComplete.NavigateUp();
					handled = true;
				}
				else if (e.KeyCode == Keys.Down)
				{
					_autoComplete.NavigateDown();
					handled = true;
				}
				else if (e.KeyCode == Keys.Enter)
				{
					_autoComplete.SelectCurrentItem();
					handled = true;
				}
				
				e.Handled = handled;
				e.SuppressKeyPress = handled;
				if (handled)
				{
					return;
				}

				// Update position...
				ShowAutoComplete();
			}

			if (e.KeyCode == Keys.Enter)
			{
				HandleEnterKey();

				e.Handled = true;
				e.SuppressKeyPress = true;
			}

			if (e.Control)
			{
				if (e.KeyCode == Keys.Space)
				{
					if (!_autoComplete.IsVisible)
					{
						ShowAutoComplete();
					}

					e.Handled = true;
					e.SuppressKeyPress = true;
				}
			}
		}

		public void InsertCommand(string text)
		{
			_commandWindow.ReplaceCurrentLineWith(text);
		}

		public void Focus()
		{
			_commandWindow.Focus();
		}
	}
}
